// -*- C++ -*-
//===----------------------------------------------------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

export {
  using ::clock_t _LIBCPP_USING_IF_EXISTS;
  using ::size_t _LIBCPP_USING_IF_EXISTS;
  using ::time_t _LIBCPP_USING_IF_EXISTS;

  using ::timespec _LIBCPP_USING_IF_EXISTS;
  using ::tm _LIBCPP_USING_IF_EXISTS;

  using ::asctime _LIBCPP_USING_IF_EXISTS;
  using ::clock _LIBCPP_USING_IF_EXISTS;
  using ::ctime _LIBCPP_USING_IF_EXISTS;
  using ::difftime _LIBCPP_USING_IF_EXISTS;
  using ::gmtime _LIBCPP_USING_IF_EXISTS;
  using ::localtime _LIBCPP_USING_IF_EXISTS;
  using ::mktime _LIBCPP_USING_IF_EXISTS;
  using ::strftime _LIBCPP_USING_IF_EXISTS;
  using ::time _LIBCPP_USING_IF_EXISTS;
  using ::timespec_get _LIBCPP_USING_IF_EXISTS;
} // export
